﻿
{ //consts
	//типы гексов									//сущности карты
	EXIT = 0;		/*выход из боя*/				UPLAND = 0; /*нагорье*/
	QUICKSAND = 1;	/*зыбучие пески*/				ROCK = 1; /*скалы*/
	BOULDER = 2;	/*валуны*/						LOWLAND = 2; /*низина*/
	PEBBLE = 3;		/*крупная галька, обломки*/		BASIN = 3; /*водоем*/
	TREE = 4;		/*деревья и кустарники*/		ROAD = 4; /*дорога*/
	DELVE = 5;		/*рытвины и колдобины*/			CRATER = 5;	 /*кратер*/
	LIGHTDELVE = 6;	/*неровная*/					STRUCTURE = 6; /*строения*/
	PLANE = 7;		/*ровная поверхность*/			WRECK = 7; /*разрушения*/
	WATER = 8;		/*вода*/						PARK = 8; /*парк*/
	BUILDING = 9;	/*здания*/						WASTELAND = 9; /*пустошь*/

	//возможная геометрия
	CURVE = 0; LINE = 1; DISK = 2; CLOSED_PATH = 3; 
	BOX = 4; BEATBOX = 5; RHOMB = 6; TRIANGLE = 7;
	
	//типы множеств
	/* set of set - множество из 0 - 5 одного из множеств из 0 - 5	*/
	SILENT = 0; CLOUD = 1; GROUP = 2; BASE = 3; MAZE = 4; MIRROR = 5; SETOFSET = 6;
	
	//типы заполнения
	FRAMED = 0; SPOTTEDFRAME = 1; SPOTTED = 2; FILLED = 3; RANDOMSPOTTED = 4;
}
{ //array-functions
	
	Array.prototype.recalcWeightsBySum = function()
	{
		var sum = 0;
		if (this && this.length>0) 
		{
			var a = this.length;
			for(var i in this) if(this.hasOwnProperty(i) && !isNaN(this[i])) sum += this[i];
			if (sum!=0) for(var i in this) if(this.hasOwnProperty(i) && !isNaN(this[i])) this[i] /= sum;
		}
	}

	Array.prototype.recalcWeightsCumulative = function()
	{
		var sum = 0;
		if (this && this.length>0) 
		{
			var a = this.length;
			var s2=0;
			for(var i in this) if(this.hasOwnProperty(i) && !isNaN(this[i])) sum += this[i];
			if (sum!=0) for(var i in this) if(this.hasOwnProperty(i) && !isNaN(this[i])) 
			{
				this[i] /= sum;
				this[i] += s2;
				s2 = this[i];
			}
			this[a-1] = 1;
		}
	}

	Array.prototype.recalcWeightsByValue = function(val_p)
	{
		if (!isNaN(val_p) && val_p!=0 && this && this.length>0) 
		{
			var a = this.length;
			for(var i in this) if(this.hasOwnProperty(i) && !isNaN(this[i])) this[i] /= val_p;
		}
	}
	
	Array.prototype.getIndexFromRangedWeightsFor = function(p_p)
	{
		var i = undefined;
		if(p_p>=0 && p_p<=1 && this.length>0)
		{
			i = 0;
			while(i<this.length && this.hasOwnProperty(i) && p_p>this[i]) i++;
		}
		return i;
	}
	
	Array.prototype.getMaxElementAndReduce = function(val_p)
	{
		var maxi={value: Number.MIN_VALUE, idx: -1};
		for(var i in this) if(this.hasOwnProperty(i) && !isNaN(this[i]) && this[i]>maxi.value)
			maxi={value: this[i], idx: i};
		
		if(maxi.idx > -1 && !isNaN(val_p)) this[maxi.idx] -= val_p;
		return maxi;
	}
}
{ //class EntityType
	function EntityType(props_obj_p)
	{
		var p = props_obj_p;
		this.name = p.name; //string
		this.min_height = p.minheight; //int
		this.max_height = p.maxheight; //int
		this.delta_height = p.deltaheight; //int
		this.topography_type = (this.max_height<=0)?LOWLAND:(this.min_height>=0)?UPLAND:PLANE; //0=upland, 2=lowland, 7=plane
		this.priority = p.priority; //приоритет при наложении гексов [0..1]
		this.size_limit = p.sizelimit;
		this.fracture_chance = p.fracture; //int - вероятн. разлома
		this.content_weights = p.content; //array - вес промилле базового гекса
		this.framing_weights = p.framing; //array - то же - обраляющего гекса
		this.geometry_weights = p.geometry; //array - то же - типа геометрии
		this.type_of_set = p.settype; // array - типы группирования
		this.fill_type = p.filltype; //тип заполнения
		
		this.content_weights.recalcWeightsCumulative();
		this.framing_weights.recalcWeightsCumulative();
		this.geometry_weights.recalcWeightsCumulative();
		this.type_of_set.recalcWeightsCumulative();
		this.fill_type.recalcWeightsCumulative();
	}
}
{ //entity types
	map_entities = [
		new EntityType
		({//горы
			name: "Нагорье", 
			minheight: 1, 
			maxheight: 8, 
			deltaheight: 2, 
			sizelimit: 16, 
			fracture: 0.2, 
			priority: 0.5,
			content: [0,600,300,30,400,200,300,0,15],
			framing: [25,100,110,25,7,15,20,1,0.02],
			geometry: [200,1,100,15,50,40,40,40],
			settype: [500,50,1,1,1],
			filltype: [0,0,10,10,1]
		}),
		new EntityType
		({//скалы
			name: "Скалы",
			minheight: -8,
			maxheight: 8,
			deltaheight: 8,
			sizelimit: 12,
			fracture: 0.4,
			priority: 0.5,
			content: [0,700,700,9,50,9,5,0,1],
			framing: [5,100,110,10,7,15,20,1,0.02],
			geometry: [150,50,5,5,30,30,30,30],
			settype: [450,50,0,2,4],
			filltype: [0,0,5,10,1]
		}),
		new EntityType
		({//низменность
			name: "Низина",
			minheight: -8,
			maxheight: -1,
			deltaheight: 2,
			sizelimit: 16, 
			fracture: 0.15,
			priority: 0.5,
			content: [0,500,400,8,90,30,5,10,10],
			framing: [10,100,110,10,15,10,5,0,0.02],
			geometry: [200,45,90,25,30,30,30,30],
			settype: [500,50,3,1,1],
			filltype: [0,0,10,10,1]
		}),
		new EntityType
		({//водоемы
			name: "Водоем",
			minheight: -8,
			maxheight: 0,
			deltaheight: 2,
			sizelimit: 6, 
			fracture: 0.04,
			priority: 0.8,
			content: [0,0,0,0,0,0,0,1,0],
			framing: [100,25,50,50,5,5,2,15,25],
			geometry: [100,3,30,15,30,30,30,30],
			settype: [600,50,1,0.5,0.5],
			filltype: [1,2,14,5,10]
		}),
		new EntityType
		({//дороги
			name: "Дорога",
			minheight: -2,
			maxheight: 5,
			deltaheight: 0.4,
			sizelimit: 5,
			fracture: 0,
			priority: 0.9,
			content: [0,1,5,0,10,200,700,0,0],
			framing: [0,150,300,150,600,350,20,0,400],
			geometry: [100,100,100,30,100,100,100,100],
			settype: [10,0,0,1,1],
			filltype: [1,1,1,50,0]
		}),
		new EntityType
		({//кратер
			name: "Кратер",
			minheight: -8,
			maxheight: -1,
			deltaheight: 5,
			sizelimit: 6,
			fracture: 0.3,
			priority: 1,
			content: [0,850,800,5,600,50,5,10,0],
			framing: [0,100,75,3,50,25,5,0,0.02],
			geometry: [10,0.5,100,15,5,0.5,0.5,0.5],
			settype: [10,100,6,0,0],
			filltype: [1,1,5,10,10]
		}),
		new EntityType
		({//building
			name: "Строение",
			minheight: 0,
			maxheight: 8,
			deltaheight: 4,
			sizelimit: 6,
			fracture: 0,
			priority: 0.9,
			content: [0,0,0,0,0,0,0,0,1],
			framing: [1,200,200,50,200,200,200,0,100],
			geometry: [20,100,100,50,200,200,200,200],
			settype: [11,50,150,100,9],
			filltype: [10,1,1,100,1]
		}),
		new EntityType
		({//развалины
			name: "Развалины",
			minheight: -4,
			maxheight: 4,
			deltaheight: 3,
			sizelimit: 6,
			fracture: 0.05,
			priority: 0.35,
			content: [0,850,500,30,60,5,0,0,5],
			framing: [1,100,200,5,40,10,1,0,15],
			geometry: [0,100,100,20,100,100,100,100],
			settype: [50,50,50,25,25],
			filltype: [1,1,1,1,1]
		}),
		new EntityType
		({//парк
			name: "Парк",
			minheight: -4,
			maxheight: 5,
			deltaheight: 4,
			sizelimit: 5,
			fracture: 0.05,
			priority: 0.5,
			content: [0,0,0,100,5,5,5,0,2],
			framing: [40,10,100,40,600,10,0,1],
			geometry: [100,25,100,50,50,75,75,50],
			settype: [600,25,25,0,3],
			filltype: [1,4,4,1,1]
		}),
		new EntityType
		({//пустошь
			name: "Пустошь",
			minheight: -4,
			maxheight: 4,
			deltaheight: 1,
			sizelimit: 20,
			fracture: 0.1,
			priority: 0.5,
			content: [200,10,30,1,300,150,50,0,3],
			framing: [100,10,5,0,20,20,20,0.01,0.02],
			geometry: [25,7,100,5,30,30,30,30],
			settype: [100,10,0,0,0],
			filltype: [0,0,1,1,1]
		})
	];
}
{ //map_types
	map_type = 
	[
		{
			name: "Пустыня", 
			entities_weight: [15,5,15,1,2,1,2,25,1,250], 
			basic_hex: [QUICKSAND,PEBBLE,DELVE,LIGHTDELVE,PLANE]
		},
		{
			name: "Пересеченная местность", 
			entities_weight: [100,50,100,5,2,20,10,20,5,100], 
			basic_hex: [BOULDER,PEBBLE,DELVE]
		},
		{
			name: "Лесной массив", 
			entities_weight: [5,0,7,15,3,1,25,7,200,35], 
			basic_hex: [QUICKSAND,PEBBLE,DELVE,LIGHTDELVE,PLANE,TREE]
		},
		{
			name: "Наводнение", 
			entities_weight: [1,5,100,250,5,3,7,7,2,1], 
			basic_hex: [QUICKSAND,BOULDER,PEBBLE,DELVE,LIGHTDELVE,PLANE]
		},
		{
			name: "Населенный пункт", 
			entities_weight: [25,0,0,15,155,1,300,10,25,30], 
			basic_hex: [BOULDER,PEBBLE,DELVE,LIGHTDELVE,PLANE,BUILDING]
		}
	];
}
{ // class Color
	function Color(red_p, green_p, blue_p)
	{
		this.red = red_p;
		this.green = green_p;
		this.blue = blue_p;
	}
}
{ //hexogen types
	hex_color = [
		new Color(0,164,212), //выход
		new Color(255,148,255), //песок
		new Color(97,97,97), //валуны, ямы
		new Color(186,186,186), //обломки, канавы
		new Color(179,225,110), //лес
		new Color(228,156,120), //галька, рытвины
		new Color(253,198,179), //наклонная поверхность
		new Color(238,235,148), //ровная поверхность
		new Color(160,130,255), // вода
		new Color(0,0,0) //здания
	];
}
{ // class Hexagon
	function Hexagon(color_obj_p, height_p, faces_arr_p)
	{
		this.color = color_obj_p;
		this.rgb = 'rgb(' + this.color.red + ',' + this.color.green + ',' + this.color.blue + ')';
		this.inverted_rgb = 'rgb(' + (255-this.color.red) + ',' + (255-this.color.green) + ',' + (255-this.color.blue) + ')';
		this.inverted_bicolor = ((this.color.red+this.color.green+this.color.blue)/3<128)?"white":"black";
		this.faces = faces_arr_p || [];
		this.height = height_p;
	}

	Hexagon.prototype.get_rgb = function()
	{
		return this.rgb;
	}

	Hexagon.prototype.get_inverted_rgb = function()
	{
		return this.inverted_rgb;
	}

	Hexagon.prototype.get_inverted_bicolor = function()
	{
		return this.inverted_bicolor;
	}

	Hexagon.prototype.get_face_in = function(p)
	{
		return ((this.faces[p] && this.faces[p]!=0)?4:0.7);
	}

	Hexagon.prototype.get_face_color = function(p)
	{
		return ((this.faces[p] && this.faces[p]!=0)?"brown":this.get_inverted_rgb());
	}
}
////////////////////////////////////////
{//class MapEntity
	function MapEntity(entity_type_p, height_lim_p, fragmentation_p)
	{
		this.hoe = []; // результирующий массив гексов сущности
		this.cet = entity_type_p; //тип выбранной сущности
		this.gt = this.cet.geometry_weights.getIndexFromRangedWeightsFor(Math.random()); //тип геометрии
		var ft1 = this.cet.fill_type.getIndexFromRangedWeightsFor(Math.random());
		var ft2 = this.cet.fill_type.getIndexFromRangedWeightsFor(Math.random());
		this.filltype = 
		{
			base: 
			{
				value: ft1,
				framed: ft1 == FRAMED,
				spottedframe: ft1 == SPOTTEDFRAME,
				spotted: ft1 == SPOTTED,
				filled: ft1 == FILLED,
				randomspotted: ft1 == RANDOMSPOTTED
			},
			nimb: 
			{
				value: ft2,
				framed: ft2 == FRAMED,
				spottedframe: ft2 == SPOTTEDFRAME,
				spotted: ft2 == SPOTTED,
				filled: ft2 == FILLED,
				randomspotted: ft2 == RANDOMSPOTTED
			}
		};
		this.es = Math.ceil(Math.random()*(this.cet.size_limit-1)+1); // размер сущности
		var ccc = this.cet.content_weights;
		var ccf = this.cet.framing_weights;
		this.nimbed = ccf && ccf.length>0 && Math.random()<0.5;
		this.hextype = 
		{
			base: hex_color[ccc.getIndexFromRangedWeightsFor(Math.random())+1],
			nimb: this.nimbed?hex_color[ccf.getIndexFromRangedWeightsFor(Math.random())+1]:undefined
		};
		
		// текущая вероятность рассеченности
		this.cfc = (fragmentation_p>0 && this.cet.fracture_chance>0)?(fragmentation_p + this.cet.fracture_chance)/2:0;
		this.damaged = 
		{
			base: Math.random()<this.cfc,
			nimb: Math.random()<this.cfc
		}
		
		// height limitations
		var thl = height_lim_p;
		var cmih = this.cet.min_height;
		var cmah = this.cet.max_height;
		this.mih = (cmih < -1*thl)?-1*thl:cmih; //минимальная высота
		this.mah = (cmah > thl)?thl:(cmah<this.mih)?this.mih:cmah; //максимальная высота
		this.mhd = this.mah - this.mih + 1;
		this.dh = Math.round(Math.random()*this.cet.delta_height); //максимальная разница высот между соседними гексами
		if(this.dh > this.mhd) this.dh = this.mhd;
		this.ih = this.dh*(2*Math.random() - 1);// initial height
		//this.ih = 0;
		this.topography = {base: PLANE, nimb: PLANE};
		switch(this.cet.topography_type)
		{
			case LOWLAND: 
				this.topography.base = (Math.random()<0.5)?PLANE:LOWLAND;
				this.topography.nimb = (Math.random()<0.5)?PLANE:LOWLAND;
			break;
			case UPLAND:
				this.topography.base = (Math.random()<0.5)?PLANE:UPLAND;
				this.topography.nimb = (Math.random()<0.5)?PLANE:UPLAND;
			break;
			default:
				var pa = Math.random(), pb = Math.random();
				this.topography.base = (pa<1/3)?PLANE:(pa<2/3)?LOWLAND:UPLAND;
				this.topography.nimb = (pb<1/3)?PLANE:(pb<2/3)?LOWLAND:UPLAND;
			break;
		}
		
		switch(this.gt) //выбор типа геометрии
		{
			case DISK: this.hoe = this.create_disk(); break;
			case BOX: this.hoe = this.create_box(); break;
			case BEATBOX: this.hoe = this.create_beatbox(); break;
			case RHOMB: this.hoe = this.create_rhomb(); break;
			case TRIANGLE: this.hoe = this.create_triangle(); break;
			case LINE: this.hoe = this.create_line();break;
			case CURVE:	this.hoe = this.create_beatbox();break;
			case CLOSED_PATH: this.hoe = this.create_closed_path(); break;
			default: this.hoe = this.create_line(); break;
		}
	}
	
	MapEntity.prototype.create_line = function()
	{
		var result = [];
		{// init data for hex
			var tb = this.es, //ширина линии
					nimbus = this.nimbed?Math.round(Math.random()*tb*1.2):0, // ширина обрамления спота
					fullnimb = nimbus*2,
					fullwidth = fullnimb + tb,
					ufw = nimbus + tb,
				tlength = Math.ceil(tb*2 + Math.random()*100), //общая длина трассы
				tdt = tb/tlength,
				m = Math.random(), //часть общей длины
				xbet = m*tlength, //длина проекции на абсциссу
				ybet = tlength*Math.pow(1 - m*m, 0.5), //длина проекции на ординату
				xaxis = xbet > ybet, //прилегает к абсциссе
				xb_ = ybet * tdt,
				yb_ = xbet * tdt,
				longest = xaxis?ybet:xbet,
				shortest = xaxis?xb_:yb_,
					clen = 0, //текущая длина трассы
					angle = Math.ceil(Math.random()*tlength), //смещение в отрезках
					minmr = Math.floor(tb/2), //мин. расст. от края основы до центра
					mingr = minmr + nimbus, //мин. расст. от края нимба до центра
				lr = Math.random()<0.5, //задом наперед?
				k = (lr?1:-1) * (xaxis?ybet/xbet:xbet/ybet), //коэф. уравнения прямой
				b = lr?0:longest,
				hy = Math.floor(ybet/2 + yb_/2);
		}
		
		for(var o=0; o<=shortest; o++) for(var i=0; i<=longest; i++)
		{
			var j = [];
			j[0] = Math.round(this.f(k,i-o,b)),
			j[1] = Math.round(this.f(k,i+o,b));
			
			for(var iii=0; iii<=1; iii++)
			{
				var ys = xaxis?j[iii]:i,
					xs = xaxis?i:j[iii];
				var dx = Math.ceil((hy-ys-1+(ys%2!=0))/2); //смещение х в зависимости от у
				
				if(this.is_to_gen(true, true, false, false, 0, 0, 0))
				{
					if(!result[ys]) result[ys]=[];
					result[ys][xs + dx] = this.calc_hex(true, {a:0, b:0}, {a:0, b:0});
				}
			}
		}
		return result;
	}
	
	MapEntity.prototype.create_triangle = function()
	{
		var result = [];
		{// init data for hex
			var tb = this.es,
				nimbus = this.nimbed?Math.round(Math.random()*tb*1.2):0, // ширина обрамления спота
				halfnimb = Math.floor(nimbus/2),
				nimbus = 2*halfnimb,
				fullnimb = 2*nimbus,
				height = Math.round(tb*(0.5 + Math.random())), //высота основы
				width = Math.round(tb*Math.random())*2, //ширина основы
				ctrb = {x: width/2, y: height/2}, //центр основы
				fh = height + fullnimb - 1, //полные высота
				lr = Math.floor(fh/2),
				fh = lr*2,
				hno = fh-nimbus,
				overturn = Math.random()<0.5,
				shift = Math.random(),
				shiftv = Math.floor((Math.random()*0.5 + 0.5)*lr),
				shifto = (shift<=0.22)?shiftv:0,
				shiftl = (shift<=0.26 && shifto==0)?shiftv:0,
				shiftr = (shift<=0.30 && shifto==0 && shiftl==0)?shiftv:0,
				sopn = shifto + nimbus,
				zaehler = 0;
		}
		
		for(var y=shifto; y<=fh; y++)
		{
			zaehler += (y>shiftv)?1:0;
			var xs = (shiftl!=0)?zaehler:0 + Math.ceil((fh-y-1+(y%2!=0))/2); //смещение х в зависимости от у
			var entw = y-zaehler; //общая ширина
			var basew = entw - nimbus;//ширина основы
			for(var x=0; x<entw; x++)
			{
				var inrange = true,
					inbox = y>=sopn && y<=hno && x>=nimbus && x<=basew,
					inframe1 = x==nimbus || x==basew-1 || y==sopn || y==hno,
					inframe2 = x==0 || x==entw-1 || y==shifto || y==fh,
					fmy = fh-y; // коорд. у с нижней стороны ромба
				
				var cmin = Math.min(Math.min(x, entw - x - 1), Math.min(y-shifto, fh-y)); //расстояние от края нимба до точки
				var minmr = Math.min(y/2 + ctrb.x, ctrb.y), //мин. расст. от края основы до центра
					mingr = minmr+nimbus+1; //мин. расст. от края нимба до центра
				var p1 = cmin/minmr; //вероятности спотов
				var p2 = cmin/nimbus;
				p1 = (Math.random()<0.7)?p1:1-p1;
				p2 = (Math.random()<0.3)?p2:1-p2;
				
				if(this.is_to_gen(inrange, inbox, inframe1, inframe2, 0.3, p1, p2))
				{
					if(overturn)
					{
						if(!result[y]) result[y]=[];
						result[y][x + xs] = this.calc_hex(inbox, {a:cmin-1, b:cmin-1}, {a:minmr, b:nimbus});
					}
					else
					{
						if(!result[fmy]) result[fmy]=[];
						result[fmy][x + xs] = this.calc_hex(inbox, {a:cmin-1, b:cmin-1}, {a:minmr, b:nimbus});
					}
				}
			}
		}
		return result;
	};
	
	MapEntity.prototype.create_rhomb = function()
	{
		var result = [];
		{// init data for hex
			var tb = this.es,
				nimbus = this.nimbed?Math.round(Math.random()*tb*1.2):0, // ширина обрамления спота
				halfnimb = Math.floor(nimbus/2),
				nimbus = 2*halfnimb,
				fullnimb = 2*nimbus,
				height = Math.round(tb*(0.5 + Math.random())), //высота основы
				width = Math.round(tb*Math.random())*2, //ширина основы
				ctrb = {x: width/2, y: height/2}, //центр основы
				fh = height + fullnimb - 1, //полные ширина и высота
				lr = Math.floor(fh/2),
				fh = lr*2;
		}
		
		for(var y=0; y<=lr; y++)
		{
			var xs = Math.ceil((lr-y-1+(y%2!=0))/2); //смещение х в зависимости от у
			var entw = width + y; //общая ширина
			var basew = entw - nimbus;//ширина основы
			for(var x=0; x<entw; x++)
			{
				var inbox = y>=nimbus && x>=nimbus && x<basew,
					inframe1 = x==nimbus || x==basew-1 || y==nimbus,
					inframe2 = x==0 || x==entw-1 || y==0,
					fmy = fh-y; // коорд. у с нижней стороны ромба
				
				var cmin = Math.min(Math.min(x, entw - x - 1), y); //расстояние от края нимба до точки
				var minmr = Math.min(y/2 + ctrb.x, ctrb.y), //мин. расст. от края основы до центра
					mingr = minmr+nimbus+1; //мин. расст. от края нимба до центра
				var p1 = cmin/minmr; //вероятности спотов
				var p2 = cmin/nimbus;
				p1 = (Math.random()<0.7)?p1:1-p1;
				p2 = (Math.random()<0.3)?p2:1-p2;
				
				if(this.is_to_gen(true, inbox, inframe1, inframe2, 0.3, p1, p2))
				{
					if(!result[y]) result[y]=[];
					result[y][x + xs] = this.calc_hex(inbox, {a:cmin-1, b:cmin-1}, {a:mingr, b:nimbus});
				}
				if(y!=lr && this.is_to_gen(true, inbox, inframe1, inframe2, 0.3, p1, p2))
				{
					if(!result[fmy]) result[fmy]=[];
					result[fmy][x + xs] = this.calc_hex(inbox, {a:cmin-1, b:cmin-1}, {a:mingr, b:nimbus});
				}
			}
		}
		return result;
	};
	
	MapEntity.prototype.create_beatbox = function()
	{
		var result = [];
		{// init data for hex
			var tb = this.es,
				nimbus = this.nimbed?Math.round(Math.random()*tb*1.2):0, // ширина обрамления спота
				fullnimb = 2*nimbus,
				width = Math.round(tb*(0.5 + Math.random())), //ширина основы
				height = Math.round(tb*(0.5 + Math.random())), //высота основы
				ctrb = {x: width/2, y: height/2}, //центр основы
				fw = width + fullnimb - 1, 
				fh = height + fullnimb - 1, //полные ширина и высота
				bb = {x1: nimbus, y1: nimbus, x2: nimbus + width - 1, y2: nimbus + height - 1}, //координаты углов основы
				minmr = (width<=height)?ctrb.x:ctrb.y, //мин. расст. от края основы до центра
				mingr = minmr+nimbus; //мин. расст. от края нимба до центра
			var lr = Math.random()<0.5;
		}
		
		for(var y=0; y<=fh; y++)
		{
			var xs = Math.ceil(lr?(y-1+(y%2!=0))/2:(fh-y-1+(y%2==0))/2);
			for(var x=0; x<=fw; x++)
			{
				var xframe = x>bb.x1 && x<bb.x2,
					yframe = y>bb.y1 && y<bb.y2,
					inbox = xframe && yframe,
					inframe1 = x==bb.x1 || x==bb.x2 || y==bb.y1 || y==bb.y2,
					inframe2 = x==0 || x==fw || y==0 || y==fh;
				
				var cmin = Math.min(Math.min(x, fw - x), Math.min(y, fh - y)); //расстояние от края нимба до точки
				var p1 = cmin/minmr; //вероятности спотов
				var p2 = cmin/nimbus;
				p1 = (Math.random()<0.7)?p1:1-p1;
				p2 = (Math.random()<0.3)?p2:1-p2;
				
				if(this.is_to_gen(true, inbox, inframe1, inframe2, 0.5, p1, p2))
				{
					if(!result[y]) result[y]=[];
					result[y][x + xs] = this.calc_hex(inbox, {a:cmin-1, b:cmin-1}, {a:mingr, b:nimbus});
				}
			}
		}
		return result;
	};
	
	MapEntity.prototype.create_box = function()
	{
		var result = [];
		{// init data for hex
			var tb = this.es,
				nimbus = this.nimbed?Math.round(Math.random()*tb*1.2):0, // ширина обрамления спота
				fullnimb = 2*nimbus,
				width = Math.round(tb*(0.5 + Math.random())), //ширина основы
				height = Math.round(tb*(0.5 + Math.random())), //высота основы
				ctrb = {x: width/2, y: height/2}, //центр основы
				fw = width + fullnimb - 1, 
				fh = height + fullnimb - 1, //полные ширина и высота
				bb = {x1: nimbus, y1: nimbus, x2: nimbus + width - 1, y2: nimbus + height - 1}, //координаты углов основы
				minmr = (width<=height)?ctrb.x:ctrb.y, //мин. расст. от края основы до центра
				mingr = minmr+nimbus; //мин. расст. от края нимба до центра
			
		}
		
		for(var y=0; y<=fh; y++)	for(var x=0; x<=fw; x++)
		{
			var xframe = x>bb.x1 && x<bb.x2,
				yframe = y>bb.y1 && y<bb.y2,
				inbox = xframe && yframe,
				inframe1 = x==bb.x1 || x==bb.x2 || y==bb.y1 || y==bb.y2,
				inframe2 = x==0 || x==fw || y==0 || y==fh;
			
			var cmin = Math.min(Math.min(x, fw - x), Math.min(y, fh - y)); //расстояние от края нимба до точки
			var p1 = cmin/minmr; //вероятности спотов
			var p2 = cmin/nimbus;
			p1 = (Math.random()<0.7)?p1:1-p1;
			p2 = (Math.random()<0.3)?p2:1-p2;
			
			if(this.is_to_gen(true, inbox, inframe1, inframe2, 0.5, p1, p2))
			{
				if(!result[y]) result[y]=[];
				result[y][x] = this.calc_hex(inbox, {a:cmin-1, b:cmin-1}, {a:mingr, b:nimbus});
			}
		}
			
		return result;
	};
	
	MapEntity.prototype.create_closed_path = function()
	{
		var result = [];
		{// init data for hex
			var tb = this.es * 4; //полный диаметр
			var wmid = this.es * 2; //полный радиус
			var dar = Math.random()*(0.6*wmid-1)+1; //радиус мертвой зоны
			var tdar = dar -1;
			var path = this.es; //половина полного радиуса
			var nimbus = this.nimbed?(wmid - dar - path)/2:0; //ширина нимба
			var n1_rad = dar + nimbus; //радиус от центра до края первого нимба
			var r2mop = n1_rad + path/2; //радиус от центра до середины траектории пути
			var path_r = n1_rad + path; //радиус от центра до на начала второго нимба
			var n2_rad = path_r + nimbus; //радиус от центра до края второго нимба
			var tn2r = n2_rad -1;
			var ddm = r2mop - dar; // расстояние от центра траектории пути до ее края
			var halfnimb = nimbus/2;
			var r2mon1 = n1_rad + halfnimb;
			var r2mon2 = n2_rad + halfnimb;
		}
		
		for(var y=0; y<tb; y++)	for(var x=0; x<tb; x++)
		{
			var dx = wmid - x; //смещение от центра
			var dy = wmid - y; //смещение от центра
			var distance = Math.pow(dx*dx + dy*dy, 0.5); //расст. от центра до точки
			var rodi = Math.round(distance)-1;
			var tdis0 = Math.abs(r2mop - distance); //расстояние от центра траектории пути до точки
			var tdis1 = ddm - tdis0; //расстояние от края траектории пути до точки
			var p1 = Math.abs(Math.round(Math.random())-tdis0/ddm); //процент удаления от центра траектории пути
			var p2 = Math.abs(Math.round(Math.random())-tdis1/nimbus); //процент удаления от края траетории пути в сторону нимба
			var tdis2 = Math.min(Math.abs(halfnimb-Math.abs(r2mon1-distance)), Math.abs(halfnimb-Math.abs(r2mon2-distance)));//расст. от края нимба до точки

			var in_range = distance>=dar && distance<=n2_rad;
			var in_path = distance>=n1_rad && distance<=path_r;
			var in_frame1 = rodi==r2mop;
			var in_frame2 = rodi==tdar || rodi==tn2r;
			
			if(this.is_to_gen(in_range, in_path, in_frame1, in_frame2, 0.33, p1, p2))
			{
				if(!result[y]) result[y]=[];
				result[y][x] = this.calc_hex(in_path, {a:tdis1-1, b:tdis2-1}, {a:ddm, b:halfnimb});
			}
		}
		
		return result;
	}
	
	MapEntity.prototype.create_disk = function()
	{
		var result = [];
		{// init data for hex
			var nimbus = this.nimbed?Math.random()*this.es*0.8:0; // ширина обрамления спота
			var tb = this.es + 2*nimbus; //полный диаметр
			var wmid = tb/2; //полный радиус
			var rwmid = Math.round(wmid)-1;
			var dtn = wmid - nimbus; //радиус базы
			var rdtn = Math.round(dtn)-1;
			var hn = nimbus/2;
		}
		
		for(var y=0; y<tb; y++)	for(var x=0; x<tb; x++)
		{
			var dx = wmid - x; //смещение от центра
			var dy = wmid - y; //смещение от центра
			var distance = Math.pow(dx*dx + dy*dy, 0.5); //расст. от центра до точки
			var rodi = Math.round(distance)-1;
			var tdis = wmid - distance; //расст. от края до точки
			var p1 = tdis/dtn;
			var p2 = tdis/nimbus;
			p1 = (Math.random()<0.7)?p1:1-p1;
			p2 = (Math.random()<0.4)?p2:1-p2;
			
			var in_range = distance<=wmid;
			var in_base = distance<=dtn;
			var in_frame1 = rodi==rdtn;
			var in_frame2 = rodi==rwmid;
			
			if(this.is_to_gen(in_range, in_base, in_frame1, in_frame2, 0.33, p1, p2))
			{
				if(!result[y]) result[y]=[];
				result[y][x] = this.calc_hex(in_base, {a:tdis-1, b:hn - Math.abs(tdis-hn)-1}, {a:dtn, b:hn});
			}
		}
		
		return result;
	};
	
	
	MapEntity.prototype.r = function(p1, p2) //дистанция меж. двумя точками с заданн. координатами
	{
		var a1 = p1.x - p2.x, 	a2 = p1.y - p2.y;
		return Math.pow(a1*a1 + a2*a2, 0.5);
	}
	
	MapEntity.prototype.f2 = function(x, p0, p1) //уровнение прямой по 2-м точкам
	{
		return p0.y + (x - p0.x)*(p1.y - p0.y)/(p1.x - p0.x);
	}
	
	MapEntity.prototype.f = function(k, p, b) //уравнение прямой с угловым коэф.
	{
		return b + k*p;
	}
	
	MapEntity.prototype.is_to_gen = function(inrange, inbase, inframe1, inframe2, sf_prob, p1, p2)
	{
		var trandom = Math.random();
		
		var in_nimbus = !inbase && inrange && this.nimbed;
		var in_frame = ((inbase && this.filltype.base.framed && inframe1)
					|| (in_nimbus && this.filltype.nimb.framed && inframe2));
		var in_spotted_frame = ((inbase && this.filltype.base.spottedframe && inframe1 && trandom<sf_prob)
					|| (in_nimbus && this.filltype.nimb.spottedframe && inframe2 && trandom<sf_prob));
		var in_spotted_area = ((inbase && this.filltype.base.spotted && trandom<p1)
					|| (in_nimbus && this.filltype.nimb.spotted && trandom<p2));
		var in_filled_area = ((inbase && this.filltype.base.filled)
					|| (in_nimbus && this.filltype.nimb.filled));
		var in_random_spotted_area = ((inbase && this.filltype.base.randomspotted)
					|| (in_nimbus && this.filltype.nimb.randomspotted));
		
		return in_frame || in_spotted_frame || in_spotted_area || in_filled_area || in_random_spotted_area;
	}
	
	MapEntity.prototype.calc_hex = function(condition_p, distance_p, radius_p)
	{
		var ccc = this.cet.content_weights;
		var ccf = this.cet.framing_weights;
		var a,b,c,d,e,f,g;
		d = this.get_random_faces();
		if(condition_p)
		{
			a = this.filltype.base.randomspotted
				?hex_color[ccc.getIndexFromRangedWeightsFor(Math.random())+1]
				:this.hextype.base;
			b = this.topography.base;
			c = this.damaged.base;
			f = distance_p.a;
			g = this.mhd/radius_p.a;
		}
		else
		{
			a = this.nimbed?
				(
					this.filltype.nimb.randomspotted
						?hex_color[ccf.getIndexFromRangedWeightsFor(Math.random())+1]
						:this.hextype.nimb
				)
				:this.hextype.base;
			b = this.topography.nimb;
			c = this.damaged.nimb;
			f = distance_p.b;
			g = this.mhd/radius_p.b;
		}
		if(this.topography.nimb == this.topography.base)
		{
			f = distance_p.a;
			g = this.mhd/(radius_p.a + radius_p.b);
		}
		e = this.calculate_height(f, g, b, c);
		return new Hexagon(a, e, d);
	};
	
	MapEntity.prototype.calculate_height = function(distance_p, height_threshold_p, topography_type_p, damaged_p)
	{
		var result = this.ih;
		var ht = (height_threshold_p<=this.dh)?height_threshold_p:this.dh;
		var k = damaged_p?Math.random():1; //рассеченность местности.
		switch(topography_type_p)
		{
			case LOWLAND:
				result = Math.floor(this.ih - ht*((distance_p + 2)*k - 1));
			break;
			case UPLAND:
				result = Math.floor(this.ih + ht*((distance_p + 2)*k - 1));
			break;
			default:
				result = Math.floor(this.ih + ht*(2*k - 1));
			break;
		}
		result = (result<this.mih)?this.mih:(result>this.mah)?this.mah:result;
		return result;
	}
	
	MapEntity.prototype.get_random_faces = function()
	{
		var result = [];
		for(var i=0;i<6;i++) result.push((Math.random()<this.cfc)?1:0);
		return result;
	}
}
{//class BattleMap
	function BattleMap(props_p)
	{
		this.hex_count_x = props_p.hex_count_x;
		this.hex_count_y = props_p.hex_count_y;
		this.cur_map_type = map_type[props_p.map_type]; //текущий тип карты
		this.fragmentation = props_p.fracture; //степень рассеченности местности 
		this.height_limitation = props_p.height_lim;
		var cmt = this.cur_map_type;
		this.basic_hex = hex_color[cmt.basic_hex[Math.floor(Math.random()*cmt.basic_hex.length)]];
		this.density = props_p.density;
		
		this.map = this.generate();
	}

	BattleMap.prototype.pushEntityToMap = function(map_p, entity_p, coord)
	{
		var result = 0;
		var hoe = entity_p.hoe;
		var rank = entity_p.cet.priority;
		for(var y in hoe) if(hoe.hasOwnProperty(y))
		{
			for(var x in hoe[y]) if(hoe[y].hasOwnProperty(x))
			{
				var cx = parseInt(coord.x) + parseInt(x);
				var cy = parseInt(coord.y) + parseInt(y);
				
				if(cx>=0 && cy>=0 && cx<this.hex_count_x && cy<this.hex_count_y)
				{
					var ch = hoe[y][x];
					if(ch)
					{
						if(map_p[cy] && map_p[cy][cx])
						{
							var vote = Math.random()<rank;
							var ht = vote?ch.color:map_p[cy][cx].color;
							var hh = Math.round((1-rank)*map_p[cy][cx].height + rank*ch.height);
							var hf = vote?ch.faces:map_p[cy][cx].faces;
							map_p[cy][cx] = new Hexagon(ht, hh, hf);
						}
						else
						{
							if(!map_p[cy]) map_p[cy]= [];
							if(!map_p[cy][cx]) map_p[cy][cx] = ch; 
						}
						result++;
					}
				}
			}
		}
		return result;
	}
	
	BattleMap.prototype.generate = function()
	{
		{// инициация данных
			var cmt = this.cur_map_type; //текущий тип карты
			var map = [];
			var t=this.basic_hex;
			var a=this.hex_count_y;
			var b=this.hex_count_x;
			var y1 = Math.floor(0.25*a), y2 = Math.floor(0.75*a);
			var x1 = Math.floor(0.25*b), x2 = Math.floor(0.75*b);
			var cmtw = cmt.entities_weight;
			var mr = this.hex_count_y*this.hex_count_x*(0.5+Math.random()*0.5)*0.01*this.density;
			var mne = Math.ceil(mr*mr); //максимум сущностей
			//alert(mne);
			if(mne<1)mne=1;
			
			var coe = []; //массив количеств сущностей на карте
			
			//определяет, сколько каких сущностей будет на карте
			for(var i=0;i<cmtw.length;i++) 
				if(cmtw[i]==0) coe.push(0);
				else coe.push(Math.floor(mne*cmtw[i]*Math.random()));
			//конец определения кол-ва сущностей
		}
		{//генерация сущностей
			var e_map = []; //карта, в которую запишутся гексы всех сущностей
			var maxe = coe.getMaxElementAndReduce(1); // выбирается сущность, которой больше всего на карте
			var condition = maxe.value>0 && maxe.idx>-1; //проверка, если сущность найдена
			while(condition)
			{
				var cet = map_entities[maxe.idx]; // тип выбранной сущности
				var current_entity = new MapEntity(cet, this.height_limitation, this.fragmentation); //ген сущности
				
				var el = // координаты на карте опорной точки сущности
				{
					x:Math.floor(Math.random()*(b + 12))-12, 
					y:Math.floor(Math.random()*(a + 12))-12
				};
				if(el.y%2==0) el.y--;
				var hexagons_token = this.pushEntityToMap(e_map, current_entity, el); //пих сущности в карту
				maxe = coe.getMaxElementAndReduce(hexagons_token);
				condition = maxe.value>0 && maxe.idx>-1;//проверка, если сущность найдена
			}
		}//конец ген. сущн.
		{// ген общей карты и выходов
			for(var y=0;y<a;y++) 
			{
				map[y] = [];
				for(var x=0;x<b;x++)
				{
					{ //разломы
					
						{// соседние гексы слева сверху
							var ltcorner=[];
							
							if(x>0) ltcorner.push({hex: map[y][x-1], s: map[y][x-1].faces[1]});
							else ltcorner.push(false);
							
							if (y>0)	
								if(y % 2 == 0)	
								{
									if(x>0) ltcorner.push({hex: map[y-1][x-1], s: map[y-1][x-1].faces[2]});
									else ltcorner.push(false);
									ltcorner.push({hex: map[y-1][x], s: map[y-1][x].faces[3]});
								}
								else 
								{
									ltcorner.push({hex: map[y-1][x], s: map[y-1][x].faces[2]});
									if(x<b-1) ltcorner.push({hex: map[y-1][x+1], s: map[y-1][x+1].faces[3]});
									else ltcorner.push(false);
								}
							else 
							{
								ltcorner.push(false);
								ltcorner.push(false);
							}
						}//конец соседние гексы
						
						{//проверка разломов
							var tfaces = [];
							var a1=4, a2=5, a3=0;
							for(var i=0;i<6;i++)
							{
								var j1 = (i==a1)?0:((i==a2)?1:((i==a3)?2:-1));
								if(j1>-1 && ltcorner[j1]) tfaces.push(ltcorner[j1].s);
								else tfaces.push((Math.random()<this.fragmentation)?1:0);
							}
						}
						
					}//конец разломы
					
					{//запись гекса
						var cht, chh, chfs;
						if(e_map[y] && e_map[y][x])
						{
							cht = e_map[y][x].color;
							chh = e_map[y][x].height;
							chfs = 
							[
								tfaces[0] || e_map[y][x].faces[0],
								e_map[y][x].faces[1],
								e_map[y][x].faces[2],
								e_map[y][x].faces[3],
								tfaces[4] || e_map[y][x].faces[4],
								tfaces[5] || e_map[y][x].faces[5]
							];
						}
						else 
						{
							cht = t;
							chh = 0;
							chfs = tfaces;
						}
						
						if( ((x==0 || x==b-1) && (y>=y1 && y<=y2)) || ((y==0 || y==a-1) && (x>=x1 && x<=x2)))
							cht = hex_color[0];
							
						map[y][x] = new Hexagon(cht, chh, chfs);
					}// конец запись гекса
				}
			}
		}
		
		return map;
	}
}
